import express, { Express } from 'express';
import cors from 'cors';
import { v4 as uuid } from 'uuid';
import { logger } from './logger';
import * as time from './time';
import Ajv from 'ajv';
import {BadInputError, ForbiddenError} from './http-errors';
import addFormats from 'ajv-formats';
import {prisma} from './db';
import path from 'path';

export enum HTTP_METHOD {
  GET = 'get',
    POST = 'post',
    DELETE = 'delete',
    PATCH = 'patch'
};

export type ApiEndpointHandler<Input, Output> = (input: Input, server: ApiServer) => Promise<Output>;

export interface WrappedApiEndpointHandler {
  method: HTTP_METHOD,
  path: string,
  handler: (req: express.Request, res: express.Response) => void;
}

const ajv = addFormats(new Ajv({}), [
    'date-time', 
    'time', 
    'date', 
    'email',  
    'hostname', 
    'ipv4', 
    'ipv6', 
    'uri', 
    'uri-reference', 
    'uuid',
    'uri-template', 
    'json-pointer', 
    'relative-json-pointer', 
    'regex'
]).addKeyword('kind').addKeyword('modifier')

export interface EndpointOptions {
  schema?: any;
  authenicated?: boolean
}


export class ApiServer {
  express: Express;
  ajv: Ajv;

  constructor() {
    this.express = express();

    this.express.use(express.json());
    this.express.use(cors());
    this.express.use(express.static(path.join(__dirname, '..', 'public')));

    this.ajv = ajv;
  }

  APIEndpointWrapper<Input, Output>(type: HTTP_METHOD, path: string, handler: ApiEndpointHandler<Input, Output>, options: EndpointOptions = {}): WrappedApiEndpointHandler {
    logger.info(`Register ${type} ${path}`);
    const hasSchema = options.schema;
    const self = this;
    return {
      method: type,
      path: path,
      handler: async (req: express.Request, res: express.Response) => {
        const start = Date.now();
        let meta = {
          id: uuid(),
          processingTime: 0,
          gameTime: time.now()
        };

        logger.info(`Req: ${req.method} ${req.path}`);
        try {
          const params = {
            params: req.params,
            body: req.body
          } as unknown;

          if(hasSchema && !this.ajv.validate(options.schema, params)) {
            if(this.ajv.errors) {
              throw new BadInputError(this.ajv.errors.map(e => e.message).join('.'));
            }
            else {
              throw new BadInputError();
            }
          }

          if(options.authenicated) {
            // lets use the token header and validate this request
            const token = await prisma.authToken.findUnique({
              where: {
                token: req.header('x-auth-token')
              }
            });

            if(!token) {
              throw new ForbiddenError('Invalid x-auth-token');
            }
          }
          
          const output: Output = await handler(params as Input, self);
          meta.processingTime = Date.now() - start;

          res.json({
            status: 'ok',
            payload: output,
            meta
          })
        }
        catch(e: any) {
          logger.error(e);
          meta.processingTime = Date.now() - start;
          res.json({
            status: 'error',
            payload: e.message,
            statusCode: e.statusCode ? e.statusCode : 500,
            meta
          });
        }
      }
    };
  }

  get<I, O>(path: string, schema: EndpointOptions, handler: ApiEndpointHandler<I, O>) {
    return this.APIEndpointWrapper<I, O>(HTTP_METHOD.GET, path, handler, schema);
  }

  post<I, O>(path: string, schema: EndpointOptions, handler: ApiEndpointHandler<I, O>) {
    return this.APIEndpointWrapper<I, O>(HTTP_METHOD.POST, path, handler, schema);
  }

  patch<I, O>(path: string, schema: EndpointOptions, handler: ApiEndpointHandler<I, O>) {
    return this.APIEndpointWrapper<I, O>(HTTP_METHOD.PATCH, path, handler, schema);
  }

  start(port: string) {
    return new Promise(res => {
      this.express.listen(port, () => {
        logger.info(`Listening on port ${port}`);
        res({port: port});
      });
    });
  }
}

export const server = new ApiServer();
