import { server } from '../../lib/server';
import { Static, Type } from '@sinclair/typebox';
import { prisma } from '../../lib/db';
import { BadInputError } from '../../lib/http-errors';
import { random } from 'lodash';
import { maxHp, maxStamina } from '../../formulas';
import {Biome, Player} from '@prisma/client';

const AccountInput = Type.Object({
  body: Type.Object({
    username: Type.String(),
    password: Type.String(),
    confirmation: Type.String()
  })
});

type AccountInputType = Static<typeof AccountInput>

export type AccountCreateType = {
  player: Player,
  biome: {
    id: string,
    biome: Biome
  }
}

export const createAccount = server.post<AccountInputType, AccountCreateType>('/v1/accounts', {
  schema: AccountInput
}, async (req) => {
  const {username, password, confirmation} = req.body;

  if(username.length < 3) {
    throw new BadInputError('Username must be at least 3 characters');
  }
  
  if(password !== confirmation) {
    throw new BadInputError('Password and confirmation did not match');
  }

  // we should get the default zone that the player will be in!

  const data = {
    username: username,
    password: password,
    pow: random(3,6),
    zest: random(3,6),
    woosh: random(3, 6),
    luck: random(3, 6),
    aha: random(3, 6),
    wow: random(3, 6),
    hp: 0,
    stamina: 0,
    zoneBiomeId: ''
  };
  


  data.hp = maxHp(1, data.zest);
  data.stamina = maxStamina(1, data.woosh);

  // create a new biome for this player!
  const biome = await prisma.zoneBiomes.create({
    data: {
      biome: 'PLAINS',
      maxSteps: 500
    }
  });

  data.zoneBiomeId = biome.id;

  const player = await prisma.player.create({ data });

  return {
    player,
    biome: {
      id: biome.id,
      biome: biome.biome
    }
  }
});
